import token
from typing import Tuple

from flynt.exceptions import FlyntException
from flynt.format import QuoteTypes

line_num = int
char_idx = int


class PyToken:
    def __init__(self, t):
        toknum, tokval, start, end, line = t
        self.toknum: int = toknum
        self.tokval: str = tokval
        self.start: Tuple[line_num, char_idx] = start
        self.end: Tuple[line_num, char_idx] = end

    def is_percent_op(self):
        return self.toknum == token.OP and self.tokval == "%"

    def is_expr_continuation_op(self):
        return (
            self.is_sq_brack_op()
            or self.is_paren_op()
            or self.is_dot_op()
            or self.is_exponentiation_op()
        )

    def is_sq_brack_op(self):
        return self.toknum == token.OP and self.tokval == "["

    def is_dot_op(self):
        return self.toknum == token.OP and self.tokval == "."

    def is_paren_op(self):
        return self.toknum == token.OP and self.tokval == "("

    def is_exponentiation_op(self):
        return self.toknum == token.OP and self.tokval == "**"

    def is_string(self):
        return self.toknum == token.STRING

    def get_quote_type(self):
        if self.toknum is not token.STRING:
            return None

        for qt in QuoteTypes.all:
            if self.tokval[: len(qt)] == qt and self.tokval[-len(qt) :] == qt:
                return qt

        if self.is_legacy_unicode_string():
            for qt in QuoteTypes.all:
                if self.tokval[1 : len(qt) + 1] == qt and self.tokval[-len(qt) :] == qt:
                    return qt

        raise FlyntException(f"Can't determine quote type of the string {self.tokval}.")

    def is_legacy_unicode_string(self):
        return self.toknum == token.STRING and self.tokval[0] == "u"

    def is_raw_string(self):
        return self.toknum == token.STRING and self.tokval[0] == "r"

    def __repr__(self):
        return f"PyToken {self.toknum} : {self.tokval}"



import io
import tokenize
import traceback
from typing import Generator

from flynt import state
from flynt.lexer.Chunk import Chunk
from flynt.lexer.PyToken import PyToken


def get_chunks(code) -> Generator[Chunk, None, None]:
    g = tokenize.tokenize(io.BytesIO(code.encode("utf-8")).readline)
    chunk = Chunk()

    try:
        for item in g:
            t = PyToken(item)
            reuse = chunk.append(t)

            if chunk.complete:

                yield chunk
                chunk = Chunk()
                if reuse:
                    reuse = chunk.append(t)
                    # assert not reuse
                    if chunk.complete:
                        yield chunk
                        chunk = Chunk()

        yield chunk
    except tokenize.TokenError as e:
        if state.verbose:
            traceback.print_exc()
            print(e)


def get_fstringify_chunks(code: str) -> Generator[Chunk, None, None]:
    """
    A generator yielding Chunks of the code where fstring can be formed.
    """
    last_concat = False

    for chunk in get_chunks(code):
        if chunk.successful and not last_concat:
            yield chunk

        if len(chunk) and chunk[-1].is_string():
            last_concat = True
        else:
            if Chunk.multiline or len(chunk) > 0:
                last_concat = False



import ast
import json


def pp_code_ast(code):
    """Pretty print code's AST to stdout."""
    tree = ast.parse(code)
    pp_ast(tree)


def _get_classname(cls):
    return cls.__class__.__name__


def ast_to_dict(node):
    """Convert an AST node to a dictionary for debugging.

    This is mainly for powering `pp_ast` (pretty printing).

    derived from `jsonify_ast` here:
    https://github.com/maligree/python-ast-explorer/blob/master/parse.py#L7

    Args:
       node (AST): the ast to turn into debug dict


    Returns a dictionary.
    """
    if not node:
        return None
    fields = {}
    for k in node._fields:
        if not hasattr(node, k):
            continue
        v = getattr(node, k)
        if isinstance(v, ast.AST):
            fields[k] = ast_to_dict(v) if v._fields else _get_classname(v)
        elif isinstance(v, list):
            fields[k] = []
            for e in v:
                fields[k].append(ast_to_dict(e))

        elif isinstance(v, (str, int, float)):
            fields[k] = v

        elif v is None:
            fields[k] = None

        else:
            fields[k] = str(v)

    return {_get_classname(node): fields}


def pp_ast(node):
    """Pretty print an AST to stdout"""
    print(json.dumps(ast_to_dict(node), indent=2))



"""This module contains global state of flynt application instance."""

verbose = False
quiet = False
aggressive = False
dry_run = False

percent_candidates = 0
percent_transforms = 0

call_candidates = 0
call_transforms = 0

invalid_conversions = 0

concat_candidates = 0
concat_changes = 0



"""Uses py3.8 AST to define a chunk of code as an AST node."""

import ast

import astor

from flynt.format import QuoteTypes


class AstChunk:
    def __init__(self, node: ast.AST):
        self.node = node

    @property
    def start_line(self):
        return self.node.lineno - 1

    @property
    def start_idx(self):
        return self.node.col_offset

    @property
    def end_idx(self):
        return self.node.end_col_offset

    @property
    def end_line(self):
        return self.node.end_lineno - 1

    @property
    def n_lines(self):
        return 1 + self.end_line - self.start_line

    @property
    def string_in_string(self):
        return False

    @property
    def quote_type(self):
        return QuoteTypes.double

    def __str__(self):
        return astor.to_source(self.node)[1:-2]

    def __repr__(self):
        return "AstChunk: " + str(self)



import ast
import sys
import string
from collections import deque

import astor

from flynt.exceptions import FlyntException, ConversionRefused


def ast_formatted_value(
    val, fmt_str: str = None, conversion=None
) -> ast.FormattedValue:

    if astor.to_source(val)[0] == "{":
        raise FlyntException(
            "values starting with '{' are better left not transformed."
        )

    format_spec = ast.JoinedStr([ast_string_node(fmt_str)]) if fmt_str else None
    conversion = -1 if conversion is None else ord(conversion.replace("!", ""))
    return ast.FormattedValue(value=val, conversion=conversion, format_spec=format_spec)


def ast_string_node(string: str) -> ast.Str:
    return ast.Str(s=string)


def matching_call(node) -> bool:
    """
    Check if an ast Node represents a "...".format() call.
    """
    return (
        isinstance(node, ast.Call)
        and hasattr(node.func, "value")
        and isinstance(node.func.value, ast.Str)
        and node.func.attr == "format"
    )


stdlib_parse = string.Formatter().parse


def joined_string(fmt_call: ast.Call) -> ast.JoinedStr:
    """ Transform a "...".format() call node into a f-string node. """

    string = fmt_call.func.value.s
    var_map = {kw.arg: kw.value for kw in fmt_call.keywords}

    for i, val in enumerate(fmt_call.args):
        var_map[i] = val

    splits = deque(stdlib_parse(string))

    seq_ctr = 0
    new_segments = []
    manual_field_ordering = False

    for raw, var_name, fmt_str, conversion in splits:
        if raw:
            new_segments.append(ast_string_node(raw))

        if var_name is None:
            continue

        if "[" in var_name:
            raise FlyntException(
                f"Skipping f-stringify of a fmt call with indexed name {var_name}"
            )

        suffix = ""
        if "." in var_name:
            idx = var_name.find(".")
            var_name, suffix = var_name[:idx], var_name[idx + 1 :]

        if var_name.isdigit():
            manual_field_ordering = True
            identifier = int(var_name)
        elif len(var_name) == 0:
            assert not manual_field_ordering
            identifier = seq_ctr
            seq_ctr += 1
        else:
            identifier = var_name

        try:
            ast_name = var_map.pop(identifier)
        except KeyError as e:
            raise ConversionRefused(
                f"A variable {identifier} is used multiple times - better not to replace it."
            ) from e

        if suffix:
            ast_name = ast.Attribute(value=ast_name, attr=suffix)
        new_segments.append(ast_formatted_value(ast_name, fmt_str, conversion))

    if var_map:
        raise FlyntException(
            f"Some variables were never used: {var_map} - skipping conversion, it's a risk of bug."
        )

    def is_literal_string(node):
        if sys.version_info < (3, 8):
            return isinstance(node, ast.Str)
        else:
            return isinstance(node, ast.Constant) and isinstance(node.value, str)

    def literal_string_value(node):
        if sys.version_info < (3, 8):
            return node.s
        else:
            return node.value

    def fix_literals(segment):
        if (
            isinstance(segment, ast.FormattedValue)
            and segment.format_spec is None
            and is_literal_string(segment.value)
        ):
            return segment.value
        return segment

    new_segments = [fix_literals(e) for e in new_segments]

    if all(is_literal_string(segment) for segment in new_segments):
        return ast.Str(
            "".join(literal_string_value(segment) for segment in new_segments)
        )

    return ast.JoinedStr(new_segments)



import ast
import sys
import token
from collections import deque
from typing import Deque

from flynt.lexer.PyToken import PyToken

REUSE = "Token was not used"


is_36 = sys.version_info.major == 3 and sys.version_info.minor == 6
if is_36:
    multiline_skip = (token.NEWLINE, 58)
    multiline_break = (57,)

    single_break = (token.NEWLINE, 57, 58)
else:
    multiline_skip = (token.NEWLINE, token.NL)
    multiline_break = (token.COMMENT,)

    single_break = (token.COMMENT, token.NEWLINE, token.NL)

single_skip = ()


class Chunk:

    skip_tokens = ()
    break_tokens = ()
    multiline = None

    @staticmethod
    def set_multiline():
        Chunk.skip_tokens = multiline_skip
        Chunk.break_tokens = multiline_break
        Chunk.multiline = True

    @staticmethod
    def set_single_line():
        Chunk.skip_tokens = single_skip
        Chunk.break_tokens = single_break
        Chunk.multiline = False

    def __init__(self, tokens=()):
        self.tokens: Deque[PyToken] = deque(tokens)
        self.complete = False

        self.is_percent_chunk = False
        self.percent_ongoing = False

        self.is_call_chunk = False
        self.successful = False

        self.string_in_string = False

    def empty_append(self, t: PyToken):
        if not t.is_string() or t.is_raw_string():
            self.complete = True

        self.tokens.append(t)

    def second_append(self, t: PyToken):
        if t.is_string():
            self.tokens[0].tokval += t.tokval
            self.tokens[0].end = t.end
        elif t.is_percent_op():
            self.tokens.append(t)
            self.is_percent_chunk = True
        elif t.is_dot_op():
            self.tokens.append(t)
            self.is_call_chunk = True
        else:
            self.tokens.append(t)
            self.complete = True

    def percent_append(self, t: PyToken):

        # todo handle all cases?
        if not self[0].is_string():
            self.complete = True
            return

        if len(self) == 2:
            self.tokens.append(t)
            if self.is_parseable:
                self.successful = True
            else:
                self.percent_ongoing = True

        else:
            if self.percent_ongoing:
                self.tokens.append(t)
                if t.is_string() and "{" not in str(self):
                    self.string_in_string = True
                if self.is_parseable:
                    self.percent_ongoing = False
                    self.successful = True
            elif t.is_expr_continuation_op():
                self.tokens.append(t)
                self.percent_ongoing = True
            else:
                self.complete = True
                self.successful = self.is_parseable
                return REUSE

    def call_append(self, t: PyToken):

        if t.is_string():
            self.string_in_string = True

        self.tokens.append(t)
        if len(self) > 3 and self.is_parseable:
            self.complete = True
            self.successful = True

    def append(self, t: PyToken):
        # stop on a comment or too long chunk
        if t.toknum in self.break_tokens:
            self.complete = True
            self.successful = self.is_parseable and (
                self.is_percent_chunk or self.is_call_chunk
            )
            return

        if len(self) > 50:
            self.complete = True
            self.successful = False
            return

        if t.toknum in self.skip_tokens:
            return

        if len(self) == 0:
            self.empty_append(t)
        elif not (self.is_call_chunk or self.is_percent_chunk):
            self.second_append(t)
        elif self.is_call_chunk:
            self.call_append(t)
        else:
            return self.percent_append(t)

    @property
    def is_parseable(self):
        if len(self.tokens) < 1:
            return False
        try:
            ast.parse(str(self))
            return True
        except SyntaxError:
            return False

    @property
    def start_line(self):
        return self.tokens[0].start[0] - 1

    @property
    def start_idx(self):
        return self.tokens[0].start[1]

    @property
    def end_idx(self):
        return self.tokens[-1].end[1]

    @property
    def end_line(self):
        return self.tokens[-1].end[0] - 1

    @property
    def n_lines(self):
        return 1 + self.end_line - self.start_line

    @property
    def is_multiline(self):
        return self.n_lines > 1

    @property
    def contains_raw_strings(self):
        return any(tok.is_raw_string() for tok in self.tokens)

    @property
    def contains_multiple_string_tokens(self):
        return sum(t.toknum == token.STRING for t in self.tokens) > 1

    @property
    def quote_type(self):
        return self.tokens[0].get_quote_type()

    def __getitem__(self, item):
        return self.tokens[item]

    def __iter__(self):
        return iter(self.tokens)

    def __len__(self):
        return len(self.tokens)

    def __str__(self):
        return " ".join(t.tokval for t in self)

    def __repr__(self):
        if self.tokens:
            return "Chunk: " + str(self)
        else:
            return "Empty Chunk"



import sys

import pytest

from flynt import process, state


@pytest.fixture()
def aggressive(monkeypatch):
    monkeypatch.setattr(state, "aggressive", True)
    yield


def test_timestamp():
    s_in = """'Timestamp: {:%Y-%m-%d %H:%M:%S}'.format(datetime.datetime.now())"""
    s_expected = """f'Timestamp: {datetime.datetime.now():%Y-%m-%d %H:%M:%S}'"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_ifexpr():
    s_in = """'%s' % (a if c else b)"""
    s_expected = """f'{a if c else b}'"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_binop():
    s_in = """'%s' % (a+b+c)"""
    s_expected = """f'{a + b + c}'"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_call():
    s_in = """'%s' % fn(var)"""
    s_expected = """f'{fn(var)}'"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_string_specific_len(aggressive):
    s_in = """'%5s' % CLASS_NAMES[labels[j]]"""
    s_expected = """f'{CLASS_NAMES[labels[j]]:5}'"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_string_in_string_single():
    s_in = """print('getlivejpg: %s: %s' % (camera['name'], errmsg))"""
    s_expected = """print(f"getlivejpg: {camera['name']}: {errmsg}")"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_percent_tuple():
    s_in = """print("%s %s " % (var+var, abc))"""
    s_expected = """print(f"{var + var} {abc} ")"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_part_of_concat():
    s_in = """print('blah{}'.format(thing) + 'blah' + otherThing + "is %f" % x)"""
    s_expected = """print(f'blah{thing}' + 'blah' + otherThing + f"is {x:f}")"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_one_string():
    s_in = """a = 'my string {}, but also {} and {}'.format(var, f, cada_bra)"""
    s_expected = """a = f'my string {var}, but also {f} and {cada_bra}'"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_nonatomic():
    s_in = """'blah{0}'.format(thing - 1)"""
    s_expected = """f'blah{thing - 1}'"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_noqa():
    s_in = """a = 'my string {}, but also {} and {}'.format(var, f, cada_bra)  # noqa: flynt"""
    s_expected = """a = 'my string {}, but also {} and {}'.format(var, f, cada_bra)  # noqa: flynt"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_noqa_other():
    s_in = """a = '%s\\n' % var  # noqa: W731, flynt"""
    s_expected = """a = '%s\\n' % var  # noqa: W731, flynt"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_multiline():
    s_in = """a = 'my string {}, but also {} and {}'.format(\nvar, \nf, \ncada_bra)"""
    s_expected = """a = f'my string {var}, but also {f} and {cada_bra}'"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_conversion():
    s_in = """a = 'my string {}, but also {!r} and {!a}'.format(var, f, cada_bra)"""
    s_expected = """a = f'my string {var}, but also {f!r} and {cada_bra!a}'"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_invalid_conversion():
    s_in = """a = 'my string {}, but also {!b} and {!a}'.format(var, f, cada_bra)"""
    s_expected = s_in

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_invalid_conversion_names():
    s_in = """a = 'my string {var}, but also {f!b}
     and {cada_bra!a}'.format(var, f, cada_bra)"""
    s_expected = s_in

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_dangerous_tuple():
    s_in = """print("%5.0f   %12.6g %12.6g %s" % (fmin_data + (step,)))"""
    s_expected = s_in

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_percent_newline():
    s_in = """a = '%s\\n' % var"""
    s_expected = """a = f'{var}\\n'"""

    s_out, count = process.fstringify_code_by_line(s_in)
    print(s_out)
    print(s_expected)
    assert s_out == s_expected


def test_format_newline():
    s_in = """a = '{}\\n'.format(var)"""
    s_expected = """a = f'{var}\\n'"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


def test_format_tab():
    s_in = """a = '{}\\t'.format(var)"""
    s_expected = """a = f'{var}\\t'"""

    s_out, count = process.fstringify_code_by_line(s_in)
    assert s_out == s_expected


indented = """
var = 5
if var % 3 == 0:
    a = "my string {}".format(var)""".strip()


def test_indented():
    s_expected = '''    a = f"my string {var}"'''
    s_out, count = process.fstringify_code_by_line(indented)

    assert count == 1
    assert s_out.split("\n")[2] == s_expected


def test_openpyxl():
    s_in = """sheet['B{}'.format(i) : 'E{}'.format(i)]"""
    s_expected = """sheet[f'B{i}' : f'E{i}']"""
    s_out, count = process.fstringify_code_by_line(s_in)

    assert count == 2
    assert s_out == s_expected


def test_str_in_str():
    s_in = """a = "beautiful numbers to follow: {}".format(" ".join(lst))"""
    s_expected = """a = f"beautiful numbers to follow: {' '.join(lst)}\""""
    s_out, count = process.fstringify_code_by_line(s_in)

    assert count == 1
    assert s_out == s_expected


def test_str_in_str_single_quote():
    s_in = """a = 'beautiful numbers to follow: {}'.format(" ".join(lst))"""
    s_expected = """a = f"beautiful numbers to follow: {' '.join(lst)}\""""
    s_out, count = process.fstringify_code_by_line(s_in)

    assert count == 1
    assert s_out == s_expected


def test_chain_fmt():
    s_in = """a = "Hello {}".format(d["a{}".format(key)])"""
    s_expected = """a = f"Hello {d[f'a{key}']}\""""
    s_out, count = process.fstringify_code_by_line(s_in)

    assert count == 1
    assert s_out == s_expected


def test_chain_fmt_3():
    s_in = """a = "Hello {}".format(d["a{}".format( d["a{}".format(key) ]) ] )"""
    s_out, count = process.fstringify_code_by_line(s_in)

    assert count == 0


code_empty_line = """
def write_row(self, xf, row, row_idx):

    attrs = {'r': '{}'.format(row_idx)}""".strip()


def test_empty_line():
    s_expected = """    attrs = {'r': f'{row_idx}'}"""
    s_out, count = process.fstringify_code_by_line(code_empty_line)

    assert count == 1
    assert s_out.split("\n")[2] == s_expected


def test_dict_perc():
    s_in = "{'r': '%s' % row_idx}"
    s_expected = """{'r': f'{row_idx}'}"""

    assert process.fstringify_code_by_line(s_in)[0] == s_expected


def test_legacy_unicode():
    s_in = """u'%s, Cadabra' % datetime.now().year"""
    s_expected = """f'{datetime.now().year}, Cadabra'"""

    assert process.fstringify_code_by_line(s_in)[0] == s_expected


def test_double_percent_no_prob():
    s_in = "{'r': '%%%s%%' % row_idx}"
    s_expected = "{'r': f'%{row_idx}%'}"

    assert process.fstringify_code_by_line(s_in)[0] == s_expected


def test_percent_dict():
    s_in = """a = '%(?)s world' % {'?': var}"""
    s_expected = """a = f'{var} world'"""

    assert process.fstringify_code_by_line(s_in)[0] == s_expected


def test_percent_dict_fmt(aggressive):
    s_in = """a = '%(?)ld world' % {'?': var}"""
    s_expected = """a = f'{int(var)} world'"""

    assert process.fstringify_code_by_line(s_in)[0] == s_expected


def test_double_percent_dict():
    s_in = """a = '%(?)s%%' % {'?': var}"""
    s_expected = """a = f'{var}%'"""

    assert process.fstringify_code_by_line(s_in)[0] == s_expected


def test_percent_dict_name():
    s_in = """a = '%(?)s world' % var"""
    s_expected = """a = f"{var['?']} world\""""

    assert process.fstringify_code_by_line(s_in)[0] == s_expected


def test_percent_dict_names():
    s_in = """a = '%(?)s %(world)s' % var"""
    s_expected = """a = f"{var['?']} {var['world']}\""""

    assert process.fstringify_code_by_line(s_in)[0] == s_expected


def test_percent_attr():
    s_in = """src_info = 'application "%s"' % srcobj.import_name"""
    s_expected = """src_info = f'application "{srcobj.import_name}"'"""

    out, count = process.fstringify_code_by_line(s_in)
    assert out == s_expected


def test_percent_dict_prefix():
    s_in = """a = '%(?)s %(world).2f' % var"""
    s_expected = """a = f"{var['?']} {var['world']:.2f}\""""

    assert process.fstringify_code_by_line(s_in)[0] == s_expected


def test_legacy_fmtspec(aggressive):
    s_in = """d = '%i' % var"""
    s_expected = """d = f'{int(var)}'"""

    out, count = process.fstringify_code_by_line(s_in)
    assert out == s_expected


def test_str_in_str_curly():
    s_in = """desired_info += ["'clusters_options' items: {}. ".format({'random_option'})]"""

    out, count = process.fstringify_code_by_line(s_in)
    assert count == 0


def test_str_in_str_methods():
    s_in = r"""string += '{} = {}\n'.format(('.').join(listKeys), json.JSONEncoder().encode(val))"""
    s_out = (
        """string += f"{'.'.join(listKeys)} = {json.JSONEncoder().encode(val)}\\n\""""
    )

    out, count = process.fstringify_code_by_line(s_in)
    assert out == s_out
    assert count > 0


def test_decimal_precision():
    s_in = """e = '%.03f' % var"""
    s_expected = """e = f'{var:.03f}'"""

    out, count = process.fstringify_code_by_line(s_in)
    assert out == s_expected


def test_width_spec(aggressive):
    s_in = "{'r': '%03f' % row_idx}"
    s_expected = """{'r': f'{row_idx:03f}'}"""

    assert process.fstringify_code_by_line(s_in)[0] == s_expected


def test_equiv_expressions_repr():
    name = "bla"  # noqa: F841

    s_in = """'Setting %20r must be uppercase.' % name"""

    out, count = process.fstringify_code_by_line(s_in)
    assert eval(out) == eval(s_in)


def test_equiv_expressions_hex():
    a = 17  # noqa: F841

    s_in = """'%.3x' % a"""

    out, count = process.fstringify_code_by_line(s_in)
    assert eval(out) == eval(s_in)


def test_equiv_expressions_s():
    name = "bla"  # noqa: F841

    s_in = """'Setting %20s must be uppercase.' % name"""

    out, count = process.fstringify_code_by_line(s_in)
    assert eval(out) == eval(s_in)


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python3.8 or higher")
def test_concat():
    s_in = """msg = a + " World\""""
    s_expected = """msg = f"{a} World\""""

    s_out, count = process.fstringify_concats(s_in)
    assert s_out == s_expected


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python3.8 or higher")
def test_concat_two_sides():
    s_in = """t = 'T is a string of value: ' + val + ' and thats great!'"""
    s_expected = """t = f"T is a string of value: {val} and thats great!\""""

    s_out, count = process.fstringify_concats(s_in)
    assert s_out == s_expected


@pytest.mark.parametrize("fmt_spec", "egdixXu")
@pytest.mark.parametrize("number", [0, 11, 0b111])
def test_integers_equivalence(number, fmt_spec):
    percent_fmt_string = f"""'Setting %{fmt_spec} must be uppercase.' % number"""
    out, count = process.fstringify_code_by_line(percent_fmt_string)

    assert eval(out) == eval(percent_fmt_string)


@pytest.mark.parametrize("fmt_spec", "egf")
@pytest.mark.parametrize("number", [3.333_333_33, 15e-44, 3.142_854])
def test_floats_equivalence(number, fmt_spec):
    percent_fmt_string = f"""'Setting %{fmt_spec} must be uppercase.' % number"""
    out, count = process.fstringify_code_by_line(percent_fmt_string)

    assert eval(out) == eval(percent_fmt_string)


@pytest.mark.parametrize("fmt_spec", [".02f", ".01e", ".04g", "05f"])
@pytest.mark.parametrize("number", [3.333_333_33, 15e-44, 3.142_854])
def test_floats_precision_equiv(number, fmt_spec):
    percent_fmt_string = f"""'Setting %{fmt_spec} must be uppercase.' % number"""
    out, count = process.fstringify_code_by_line(percent_fmt_string)

    assert eval(out) == eval(percent_fmt_string)



import ast
import copy
import traceback
from typing import Tuple

import astor

from flynt import state
from flynt.exceptions import FlyntException, ConversionRefused
from flynt.format import QuoteTypes, set_quote_type
from flynt.transform.FstringifyTransformer import fstringify_node


def transform_chunk(
    code: str, quote_type: str = QuoteTypes.triple_double
) -> Tuple[str, bool]:
    """Convert a block of code to an f-string

    Args:
        code: The code to convert.
        quote_type: the quote type to use for the transformed result

    Returns:
       Tuple: resulting code, boolean: was it changed?
    """

    try:
        tree = ast.parse(code)
        converted, changed, str_in_str = fstringify_node(copy.deepcopy(tree))
    except (SyntaxError, FlyntException, Exception) as e:
        if state.verbose:
            if isinstance(e, ConversionRefused):
                print(f"Not converting code '{code}': {e}")
            else:
                print(f"Exception {e} during conversion of code '{code}'")
                traceback.print_exc()
        state.invalid_conversions += 1
        return code, False
    else:
        if changed:
            new_code = astor.to_source(converted)
            new_code = new_code.strip()
            if str_in_str and quote_type == QuoteTypes.single:
                quote_type = QuoteTypes.double
            new_code = set_quote_type(new_code, quote_type)
            new_code = new_code.replace("\n", "\\n")
            new_code = new_code.replace("\t", "\\t")
            try:
                ast.parse(new_code)
            except Exception as e:
                if state.verbose:
                    print(
                        f"Failed to parse transformed code '{new_code}' given original '{code}'"
                    )
                    print(e)
                    traceback.print_exc()
                state.invalid_conversions += 1
                return code, False
            else:
                return new_code, changed

        return code, False



import ast

from flynt.exceptions import StringEmbeddingTooDeep


class SinSDetector(ast.NodeVisitor):
    def __init__(self, maxdepth):

        self.sns_depth = 0
        self.maxdepth = maxdepth

    def visit_JoinedStr(self, node):
        self.sns_depth += 1
        if self.sns_depth > self.maxdepth:
            raise StringEmbeddingTooDeep

        self.generic_visit(node)
        self.sns_depth -= 1

    def visit_FormattedValue(self, node):
        self.visit(node.value)


def check_sns_depth(node, limit=1):
    d = SinSDetector(limit)
    try:
        d.visit(node)
    except StringEmbeddingTooDeep:
        return False
    else:
        return True



"""This module deals with quote types for strings.

It includes checking the quote type and changing it."""

import io
import re
import tokenize

lonely_quote = re.compile(r"(?<!\\)\"")


class QuoteTypes:
    single = "'"
    double = '"'
    triple_single = "'''"
    triple_double = '"""'
    all = [triple_double, triple_single, single, double]


def get_quote_type(code: str):
    from flynt.lexer.PyToken import PyToken

    g = tokenize.tokenize(io.BytesIO(code.encode("utf-8")).readline)
    next(g)
    token = PyToken(next(g))

    return token.get_quote_type()


def remove_quotes(code: str):
    quote_type = get_quote_type(code)
    return code[len(quote_type) : -len(quote_type)]


def set_quote_type(code: str, quote_type: str):
    if code[0] == "f":
        prefix, body = "f", remove_quotes(code[1:])
    else:
        prefix, body = "", remove_quotes(code)
    if quote_type in (QuoteTypes.single, QuoteTypes.triple_double):
        if body[-2:] == '\\"':
            body = body[:-2] + '"'
    elif quote_type is QuoteTypes.double:
        body = lonely_quote.sub('\\"', body)

    return prefix + quote_type + body + quote_type



import ast
from typing import List, Tuple

import astor

from flynt.exceptions import FlyntException
from flynt.format import QuoteTypes, set_quote_type
from flynt.string_concat.candidates import is_string_concat
from flynt.string_concat.string_in_string import check_sns_depth
from flynt.linting.fstr_lint import FstrInliner


def ast_formatted_value(
    val, fmt_str: str = None, conversion=None
) -> ast.FormattedValue:
    if isinstance(val, ast.FormattedValue):
        return val

    if astor.to_source(val)[0] == "{":
        raise FlyntException(
            "values starting with '{' are better left not transformed."
        )

    if fmt_str:
        format_spec = ast.JoinedStr([ast_string_node(fmt_str.replace(":", ""))])
    else:
        format_spec = None

    conversion = -1 if conversion is None else ord(conversion.replace("!", ""))
    return ast.FormattedValue(value=val, conversion=conversion, format_spec=format_spec)


def ast_string_node(string: str) -> ast.Str:
    return ast.Str(s=string)


def unpack_binop(node: ast.BinOp) -> List[ast.AST]:
    assert isinstance(node, ast.BinOp)
    result = []

    if isinstance(node.left, ast.BinOp):
        result.extend(unpack_binop(node.left))
    else:
        result.append(node.left)

    if isinstance(node.right, ast.BinOp):
        result.extend(unpack_binop(node.right))
    else:
        result.append(node.right)

    return result


class ConcatTransformer(ast.NodeTransformer):
    def __init__(self):
        super().__init__()
        self.counter = 0

    def visit_BinOp(self, node: ast.BinOp):
        """
        Transforms a string concat to an f-string
        """
        if not is_string_concat(node):
            return self.generic_visit(node)
        self.counter += 1
        left, right = node.left, node.right
        left = self.visit(left)
        right = self.visit(right)

        if not check_sns_depth(left) or not check_sns_depth(right):
            node.left = left
            node.right = right
            return node

        parts = []
        for p in [left, right]:
            if isinstance(p, ast.JoinedStr):
                parts += p.values
            else:
                parts.append(p)

        segments = []
        for p in parts:
            if isinstance(p, ast.Constant):
                segments.append(ast_string_node(p.value))
            else:
                segments.append(ast_formatted_value(p))

        return ast.JoinedStr(segments)


def transform_concat(code: str, *args, **kwargs) -> Tuple[str, bool]:
    tree = ast.parse(f"({code})")

    ft = ConcatTransformer()
    ft.visit(tree)
    il = FstrInliner()
    il.visit(tree)

    new_code = astor.to_source(tree)
    if new_code[-1] == "\n":
        new_code = new_code[:-1]

    new_code = new_code.replace("\n", "\\n")
    if new_code[:4] == 'f"""':
        new_code = set_quote_type(new_code, QuoteTypes.double)

    return new_code, ft.counter > 0



"""This module parses the command line arguments and passes them to flynt.api.fstringify."""

import argparse
import sys

from flynt.api import fstringify
from flynt import state
from flynt import __version__


def main():

    parser = argparse.ArgumentParser(
        description=f"flynt v.{__version__}", add_help=True, epilog=__doc__
    )

    verbosity_group = parser.add_mutually_exclusive_group()
    verbosity_group.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="run with verbose output",
        default=False,
    )
    verbosity_group.add_argument(
        "-q", "--quiet", action="store_true", help="run without output", default=False
    )

    multiline_group = parser.add_mutually_exclusive_group()
    multiline_group.add_argument(
        "--no_multiline", action="store_true", help=argparse.SUPPRESS, default=False
    )

    multiline_group.add_argument(
        "--no-multiline",
        action="store_true",
        help="convert only single line expressions",
        default=False,
    )

    multiline_group.add_argument(
        "--line_length", action="store", help=argparse.SUPPRESS, default=88
    )

    multiline_group.add_argument(
        "-ll",
        "--line-length",
        action="store",
        help="for expressions spanning multiple lines, convert only if "
        "the resulting single line will fit into the line length limit. "
        "Default value is 88 characters.",
        default=88,
    )

    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        default=False,
        help="Do not change the files in-place and print the diff instead. "
        "Note that this must be used in conjunction with '--fail-on-change' when "
        "used for linting purposes.",
    )

    parser.add_argument(
        "-tc",
        "--transform-concats",
        action="store_true",
        default=False,
        help="Replace string concatenations (defined as + operations involving string literals) "
        "with f-strings. Available only if flynt is installed with 3.8+ interpreter.",
    )

    parser.add_argument(
        "-f",
        "--fail-on-change",
        action="store_true",
        default=False,
        help="Fail when changing files (for linting purposes)",
    )

    parser.add_argument(
        "-a",
        "--aggressive",
        action="store_true",
        default=False,
        help="Include conversions with potentially changed behavior.",
    )

    parser.add_argument(
        "-e",
        "--exclude",
        action="store",
        nargs="+",
        help="ignore files with given strings in it's absolute path.",
    )

    parser.add_argument(
        "src", action="store", nargs="*", help="source file(s) or directory"
    )

    parser.add_argument(
        "--version",
        action="store_true",
        default=False,
        help="Print the current version number and exit.",
    )
    args = parser.parse_args()

    if args.version:
        print(__version__)
        sys.exit(0)
    elif not args.src:
        print("flynt: error: the following arguments are required: src")
        parser.print_usage()
        sys.exit(1)

    print(f"Running flynt v.{__version__}")
    if args.dry_run:
        print("Running flynt in dry-run mode. No files will be changed.")

    if args.transform_concats and sys.version_info < (3, 8):
        raise Exception(
            f"""Transforming string concatenations is only possible with flynt 
                installed to a python3.8+ interpreter. Currently using {sys.version_info}."""
        )

    state.aggressive = args.aggressive
    state.verbose = args.verbose
    state.quiet = args.quiet
    state.dry_run = args.dry_run

    return fstringify(
        args.src,
        excluded_files_or_paths=args.exclude,
        multiline=not args.no_multiline,
        len_limit=int(args.line_length),
        fail_on_changes=args.fail_on_change,
        transform_concat=args.transform_concats,
    )



import ast
import os
import sys
import time
import traceback
from typing import Tuple, List
from difflib import unified_diff

import astor

from flynt import state
from flynt.cli_messages import farewell_message
from flynt.process import fstringify_code_by_line, fstringify_concats

blacklist = {".tox", "venv", "site-packages", ".eggs"}


def _fstringify_file(
    filename, multiline, len_limit, transform_concat=False
) -> Tuple[bool, int, int, int]:
    """
    :return: tuple: (changes_made, n_changes,
    length of original code, length of new code)
    """

    with open(filename, encoding="utf-8") as f:
        contents = f.read()

    def default_result():
        return False, 0, len(contents), len(contents)

    try:
        ast_before = ast.parse(contents)
    except SyntaxError:
        if not state.quiet:
            print(f"Can't parse {filename} as a python file.")
        return default_result()

    try:
        new_code, changes = fstringify_code_by_line(
            contents, multiline=multiline, len_limit=len_limit
        )
        if transform_concat:
            new_code, concat_changes = fstringify_concats(
                new_code, multiline=multiline, len_limit=len_limit
            )
            changes += concat_changes
            state.concat_changes += concat_changes

    except Exception as e:
        if not state.quiet:
            if str(e):
                msg = str(e)
            else:
                msg = e.__class__.__name__
            print(f"Skipping fstrings transform of file {filename} due to {msg}.")
            if state.verbose:
                traceback.print_exc()

        return default_result()

    if new_code == contents:
        return default_result()

    try:
        ast_after = ast.parse(new_code)
    except SyntaxError:
        print(f"Faulty result during conversion on {filename} - skipping.")
        if state.verbose:
            print(new_code)
            traceback.print_exc()
        return default_result()

    if not len(ast_before.body) == len(ast_after.body):
        print(
            f"Faulty result during conversion on {filename}: "
            f"statement count has changed, which is not intended - skipping."
        )
        return default_result()

    if state.dry_run:
        for l in unified_diff(
            contents.split("\n"), new_code.split("\n"), fromfile=filename
        ):
            print(l)
    else:
        with open(filename, "w", encoding="utf-8", newline="") as f:
            f.write(new_code)

    return True, changes, len(contents), len(new_code)


def fstringify_files(files, multiline, len_limit, transform_concat):
    changed_files = 0
    total_charcount_original = 0
    total_charcount_new = 0
    total_expressions = 0
    start_time = time.time()
    for path in files:
        (
            changed,
            count_expressions,
            charcount_original,
            charcount_new,
        ) = _fstringify_file(path, multiline, len_limit, transform_concat)
        if changed:
            changed_files += 1
            total_expressions += count_expressions
        total_charcount_original += charcount_original
        total_charcount_new += charcount_new

        if state.verbose:
            status = "modified" if count_expressions else "no change"
            print(f"fstringifying {path}...{status}")
    total_time = time.time() - start_time

    if not state.quiet:
        _print_report(
            changed_files,
            total_charcount_new,
            total_charcount_original,
            total_expressions,
            total_time,
        )

    return changed_files


def _print_report(
    changed_files, total_cc_new, total_cc_original, total_expr, total_time
):
    print("\nFlynt run has finished. Stats:")
    print(f"\nExecution time:                            {total_time:.3f}s")
    print(f"Files modified:                            {changed_files}")
    if changed_files:
        cc_reduction = total_cc_original - total_cc_new
        cc_percent_reduction = cc_reduction / total_cc_original
        print(
            f"Character count reduction:                 {cc_reduction} ({cc_percent_reduction:.2%})\n"
        )

        print("Per expression type:")
        if state.percent_candidates:
            percent_fraction = state.percent_transforms / state.percent_candidates
            print(
                f"Old style (`%`) expressions attempted:     {state.percent_transforms}/"
                f"{state.percent_candidates} ({percent_fraction:.1%})"
            )
        else:
            print("No old style (`%`) expressions attempted.")

        if state.call_candidates:
            print(
                f"`.format(...)` calls attempted:            {state.call_transforms}/"
                f"{state.call_candidates} ({state.call_transforms / state.call_candidates:.1%})"
            )
        else:
            print("No `.format(...)` calls attempted.")

        if state.concat_candidates:
            print(
                f"String concatenations attempted:           {state.concat_changes}/"
                f"{state.concat_candidates} ({state.concat_changes / state.concat_candidates:.1%})"
            )
        else:
            print("No concatenations attempted.")

        print(f"F-string expressions created:              {total_expr}")

        if state.invalid_conversions:
            print(
                f"Out of all attempted transforms, {state.invalid_conversions} resulted in errors."
            )
            print("To find out specific error messages, use --verbose flag.")

    print("\n" + ("_-_." * 25))
    print(farewell_message)


def fstringify(
    files_or_paths,
    multiline,
    len_limit,
    fail_on_changes=False,
    transform_concat=False,
    excluded_files_or_paths=None,
):
    """ determine if a directory or a single file was passed, and f-stringify it."""

    files = _resolve_files(files_or_paths, excluded_files_or_paths)

    status = fstringify_files(
        files,
        multiline=multiline,
        len_limit=len_limit,
        transform_concat=transform_concat,
    )

    if fail_on_changes:
        return status
    else:
        return 0


def _resolve_files(files_or_paths, excluded_files_or_paths) -> List[str]:
    """Resolve relative paths and directory names into a list of absolute paths to python files."""

    files = []
    _blacklist = blacklist.copy()
    if excluded_files_or_paths is not None:
        _blacklist.update(set(excluded_files_or_paths))

    for file_or_path in files_or_paths:

        abs_path = os.path.abspath(file_or_path)

        if not os.path.exists(abs_path):
            print(f"`{file_or_path}` not found")
            sys.exit(1)

        if os.path.isdir(abs_path):
            for folder, filename in astor.code_to_ast.find_py_files(abs_path):
                files.append(os.path.join(folder, filename))
        else:
            files.append(abs_path)

    files = [f for f in files if all(b not in f for b in _blacklist)]
    return files



import math
import re
import traceback
from typing import Callable, Tuple
import string

from flynt import lexer, state
from flynt.exceptions import FlyntException
from flynt.format import QuoteTypes as qt
from flynt.format import get_quote_type
from flynt.lexer import split
from flynt.transform.transform import transform_chunk
from flynt.string_concat import concat_candidates, transform_concat

noqa_regex = re.compile("#[ ]*noqa.*flynt")


class JoinTransformer:
    """JoinTransformer fills up the resulting code by tracking
    the last line number and char index. Failed transformations do not need to do anything -
    not adding results is safe, as original code will be filled in."""

    def __init__(
        self,
        code: str,
        len_limit: int,
        candidates_iter_factory: Callable,
        transform_func: Callable,
    ):

        if len_limit is None:
            len_limit = math.inf

        self.len_limit = len_limit
        self.candidates_iter = candidates_iter_factory(code)
        self.transform_func = transform_func
        self.src_lines = code.split("\n")

        self.results = []
        self.count_expressions = 0

        self.last_line = 0
        self.last_idx = 0
        self.used_up = False

    def fstringify_code_by_line(self):
        assert not self.used_up, "Tried to use JT twice."
        for chunk in self.candidates_iter:
            self.fill_up_to(chunk)
            self.try_chunk(chunk)

        self.add_rest()
        self.used_up = True
        return "".join(self.results)[:-1], self.count_expressions

    def fill_up_to(self, chunk):
        start_line, start_idx, _ = (chunk.start_line, chunk.start_idx, chunk.end_idx)
        if start_line == self.last_line:
            self.results.append(
                self.src_lines[self.last_line][self.last_idx : start_idx]
            )
        else:
            self.results.append(self.src_lines[self.last_line][self.last_idx :] + "\n")
            self.last_line += 1
            line = self.src_lines[start_line]

            self.fill_up_to_line(start_line)
            self.results.append(line[:start_idx])

        self.last_idx = start_idx

    def fill_up_to_line(self, start_line):
        while self.last_line < start_line:
            self.results.append(self.src_lines[self.last_line] + "\n")
            self.last_line += 1

    def try_chunk(self, chunk):

        for line in self.src_lines[chunk.start_line : chunk.start_line + chunk.n_lines]:
            if noqa_regex.findall(line):
                # user does not wish for this line to be converted.
                return

        try:
            quote_type = (
                qt.double
                if chunk.string_in_string and chunk.n_lines == 1
                else chunk.quote_type
            )
        except FlyntException as e:
            if state.verbose:
                print(f"Exception {e} during conversion of code '{str(chunk)}'")
                traceback.print_exc()

        else:
            converted, changed = self.transform_func(str(chunk), quote_type=quote_type)
            if changed:
                contract_lines = chunk.n_lines - 1
                if contract_lines == 0:
                    line = self.src_lines[chunk.start_line]
                    rest = line[chunk.end_idx :]
                else:
                    next_line = self.src_lines[chunk.start_line + contract_lines]
                    rest = next_line[chunk.end_idx :]
                self.maybe_replace(chunk, contract_lines, converted, rest)

    def maybe_replace(self, chunk, contract_lines, converted, rest):
        if contract_lines:
            if get_quote_type(str(chunk)) in (qt.triple_double, qt.triple_single):
                lines = converted.split("\\n")
                lines[-1] += rest
                lines_fit = all(
                    len(l) <= self.len_limit - chunk.start_idx for l in lines
                )
                converted = converted.replace("\\n", "\n")
            else:
                lines_fit = (
                    len("".join([converted, rest])) <= self.len_limit - chunk.start_idx
                )

        else:
            lines_fit = True
        if not contract_lines or lines_fit:

            self.results.append(converted)
            self.count_expressions += 1
            self.last_line += contract_lines
            self.last_idx = chunk.end_idx

            # remove redundant parenthesis
            if len(self.results) < 2 or not self.results[-2]:
                return
            elif len(self.src_lines[self.last_line]) == self.last_idx:
                return

            if (
                self.results[-2][-1] == "("
                and self.src_lines[self.last_line][self.last_idx] == ")"
            ):
                for char in reversed(self.results[-2][:-1]):
                    if char in string.whitespace:
                        continue
                    elif char in "(=[+*":
                        break
                    else:
                        return

                self.results[-2] = self.results[-2][:-1]
                self.last_idx += 1

    def add_rest(self):
        self.results.append(self.src_lines[self.last_line][self.last_idx :] + "\n")
        self.last_line += 1

        while len(self.src_lines) > self.last_line:
            self.results.append(self.src_lines[self.last_line] + "\n")
            self.last_line += 1


def fstringify_code_by_line(code: str, multiline=True, len_limit=88) -> Tuple[str, int]:
    """ returns fstringified version of the code and amount of lines edited."""
    return _transform_code(
        code, split.get_fstringify_chunks, transform_chunk, multiline, len_limit
    )


def fstringify_concats(code: str, multiline=True, len_limit=88) -> Tuple[str, int]:
    """ replace string literal concatenations with f-string expressions. """
    return _transform_code(
        code, concat_candidates, transform_concat, multiline, len_limit
    )


def _transform_code(
    code: str, candidates_iter_factory, transform_func, multiline, len_limit
) -> Tuple[str, int]:
    """ returns fstringified version of the code and amount of lines edited."""

    len_limit = _multiline_settings(len_limit, multiline)
    jt = JoinTransformer(code, len_limit, candidates_iter_factory, transform_func)
    return jt.fstringify_code_by_line()


def _multiline_settings(len_limit, multiline):
    if not multiline:
        len_limit = 0
        lexer.set_single_line()
    else:
        lexer.set_multiline()
    return len_limit



import ast
from typing import Tuple

from flynt import state
from flynt.transform.format_call_transforms import joined_string, matching_call
from flynt.transform.percent_transformer import transform_binop, supported_operands
from flynt.linting.fstr_lint import FstrInliner


class FstringifyTransformer(ast.NodeTransformer):
    def __init__(self):
        super().__init__()
        self.counter = 0
        self.string_in_string = False

    def visit_Call(self, node: ast.Call):
        """
        Convert `ast.Call` to `ast.JoinedStr` f-string
        """

        match = matching_call(node)

        if match:
            state.call_candidates += 1

            # bail in these edge cases...
            if any(isinstance(arg, ast.Starred) for arg in node.args):
                return node

            result_node = joined_string(node)
            self.visit(result_node)
            self.counter += 1
            state.call_transforms += 1
            return result_node

        return node

    def visit_BinOp(self, node):
        """Convert `ast.BinOp` to `ast.JoinedStr` f-string

        Currently only if a string literal `ast.Str` is on the left side of the `%`
        and one of `ast.Tuple`, `ast.Name`, `ast.Dict` is on the right

        Args:
            node (ast.BinOp): The node to convert to a f-string

        Returns ast.JoinedStr (f-string)
        """

        percent_stringify = (
            isinstance(node.left, ast.Str)
            and isinstance(node.op, ast.Mod)
            and isinstance(
                node.right,
                tuple([ast.Tuple, ast.Dict] + supported_operands),
            )
        )

        if percent_stringify:
            state.percent_candidates += 1

            # bail in these edge cases...
            no_good = ["}", "{"]
            for ng in no_good:
                if ng in node.left.s:
                    return node
            for ch in ast.walk(node.right):
                # f-string expression part cannot include a backslash
                if isinstance(ch, ast.Str) and (
                    any(
                        map(
                            lambda x: x in ch.s,
                            ("\n", "\t", "\r", "'", '"', "%s", "%%"),
                        )
                    )
                    or "\\" in ch.s
                ):
                    return node

            result_node, str_in_str = transform_binop(node)
            self.string_in_string = str_in_str
            self.counter += 1
            state.percent_transforms += 1
            return result_node

        return node


def fstringify_node(node) -> Tuple[ast.AST, bool, bool]:
    ft = FstringifyTransformer()
    result = ft.visit(node)
    il = FstrInliner()
    il.visit(result)

    return result, ft.counter > 0, ft.string_in_string



import ast
import re
from collections import deque

from flynt.exceptions import FlyntException, ConversionRefused
from flynt.transform.format_call_transforms import ast_formatted_value, ast_string_node
from flynt import state

FORMATS = "diouxXeEfFgGcrsa"

FORMAT_GROUP = f"[hlL]?[{FORMATS}]"
FORMAT_GROUP_MATCH = f"[hlL]?([{FORMATS}])"

PREFIX_GROUP = "[0-9]*[.]?[0-9]*"

DICT_PATTERN = re.compile(rf"(%\([^)]+\){PREFIX_GROUP}{FORMAT_GROUP})")
SPLIT_DICT_PATTERN = re.compile(rf"%\(([^)]+)\)({PREFIX_GROUP}){FORMAT_GROUP_MATCH}")
VAR_KEY_PATTERN = re.compile(
    f"%({PREFIX_GROUP}){FORMAT_GROUP_MATCH}"
)  # specs at https://docs.python.org/3/library/stdtypes.html#string-formatting
obsolete_specifiers = "hlL"

translate_conversion_types = {"i": "d", "u": "d"}
conversion_methods = {"r": "!r", "a": "!a", "s": None}
integer_specificers = "dxXob"


def formatted_value(fmt_prefix, fmt_spec, val):
    if fmt_spec in integer_specificers:
        fmt_prefix = fmt_prefix.replace(".", "0")

    if fmt_spec in conversion_methods:
        if not state.aggressive and fmt_prefix:
            raise FlyntException(
                "Default text alignment has changed between percent fmt and fstrings. "
                "Proceeding would result in changed code behaviour."
            )
        fv = ast_formatted_value(
            val, fmt_str=fmt_prefix, conversion=conversion_methods[fmt_spec]
        )
    else:
        fmt_spec = translate_conversion_types.get(fmt_spec, fmt_spec)
        if fmt_spec == "d":
            if state.aggressive:
                val = ast.Call(
                    func=ast.Name(id="int", ctx=ast.Load()), args=[val], keywords={}
                )
                fmt_spec = ""
            else:
                raise FlyntException(
                    "Skipping %d formatting - fstrings behave differently from % formatting."
                )
        fv = ast_formatted_value(val, fmt_str=fmt_prefix + fmt_spec)
    return fv


def transform_dict(node):
    """Convert a `BinOp` `%` formatted str with a name representing a Dict on the right to an f-string.

    Takes an ast.BinOp representing `"1. %(key1)s 2. %(key2)s" % mydict`
    and converted it to a ast.JoinedStr representing `f"1. {mydict['key1']} 2. {mydict['key2']}"`

    Args:
       node (ast.BinOp): The node to convert to a f-string

    Returns ast.JoinedStr (f-string)
    """

    format_str = node.left.s
    matches = DICT_PATTERN.findall(format_str)
    spec = []
    for idx, m in enumerate(matches):
        _, var_key, prefix, fmt_str, _ = SPLIT_DICT_PATTERN.split(m)
        if not var_key:
            raise FlyntException("could not find dict key")
        spec.append((prefix, var_key, fmt_str))

    # build result node
    segments = []
    spec.reverse()
    blocks = DICT_PATTERN.split(format_str)

    mapping = {}
    if isinstance(node.right, ast.Dict):
        for k, v in zip(node.right.keys, node.right.values):
            mapping[str(ast.literal_eval(k))] = v

        def make_fv(key: str):
            return mapping.pop(key)

    else:

        def make_fv(key: str):
            return ast.Subscript(
                value=node.right, slice=ast.Index(value=ast.Str(s=key))
            )

    for block in blocks:
        # if this block matches a %(arg)s pattern then inject f-string instead
        if DICT_PATTERN.match(block):
            prefix, var_key, fmt_str = spec.pop()
            fv = formatted_value(prefix, fmt_str, make_fv(var_key))
            segments.append(fv)
        else:
            # no match means it's just a literal string
            segments.append(ast.Str(s=block.replace("%%", "%")))

    if mapping:
        raise FlyntException(
            "Not all keys were matched - either a flynt error or original code error."
        )

    return ast.JoinedStr(segments)


def transform_tuple(node):
    """Convert a `BinOp` `%` formatted str with a tuple on the right to an f-string.

    Takes an ast.BinOp representing `"1. %s 2. %s" % (a, b)`
    and converted it to a ast.JoinedStr representing `f"1. {a} 2. {b}"`

    Args:
       node (ast.BinOp): The node to convert to a f-string

    Returns ast.JoinedStr (f-string)
    """

    format_str = node.left.s
    matches = VAR_KEY_PATTERN.findall(format_str)

    if len(node.right.elts) != len(matches):
        raise ConversionRefused("This expression involves tuple unpacking.")

    str_vars = deque(node.right.elts)

    segments = []
    blocks = deque(VAR_KEY_PATTERN.split(format_str))
    segments.append(ast_string_node(blocks.popleft().replace("%%", "%")))

    while len(blocks) > 0:

        fmt_prefix = blocks.popleft()
        fmt_spec = blocks.popleft()
        val = str_vars.popleft()

        fv = formatted_value(fmt_prefix, fmt_spec, val)

        segments.append(fv)
        segments.append(ast_string_node(blocks.popleft().replace("%%", "%")))

    return ast.JoinedStr(segments)


def transform_generic(node):
    """Convert a `BinOp` `%` formatted str with a unknown name on the `node.right` to an f-string.

    When `node.right` is a Name since we don't know if it's a single var or a dict so we sniff the string.

    Sniffs the left string for Dict style usage
    e.g. `"val: %(key_name1)s val2: %(key_name2)s" % some_dict`

    else (e.g. `"val: %s" % some_var`):
    Borrow the core logic by injecting the name into a ast.Tuple

    Returns ast.JoinedStr (f-string), bool: str-in-str
    """

    has_dict_str_format = DICT_PATTERN.findall(node.left.s)
    if has_dict_str_format:
        return transform_dict(node), True

    # if it's just a name then pretend it's tuple to use that code
    node.right = ast.Tuple(elts=[node.right])
    return transform_tuple(node), False


supported_operands = [
    ast.Name,
    ast.Attribute,
    ast.Str,
    ast.Subscript,
    ast.Call,
    ast.BinOp,
    ast.IfExp,
]


def transform_binop(node):
    if isinstance(
        node.right,
        tuple(supported_operands),
    ):
        return transform_generic(node)

    elif isinstance(node.right, ast.Tuple):
        return transform_tuple(node), False

    elif isinstance(node.right, ast.Dict):
        # todo adapt transform dict to Dict literal
        return transform_dict(node), False

    raise ConversionRefused(f"Unsupported `node.right` class: {type(node.right)}")



"""`flynt` is a command line tool to automatically convert a project's Python code
from old "%-formatted" and .format(...) strings into Python 3.6+'s f-strings.
Learn more about f-strings at https://www.python.org/dev/peps/pep-0498/"""

__version__ = "0.60"

from flynt.cli import main



import ast
from typing import List

from flynt.ast_chunk import AstChunk


class FstringFinder(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.victims: List[AstChunk] = []

    def visit_JoinedStr(self, node: ast.BinOp):
        """
        Finds all nodes that are string concatenations with a literal
        """
        self.victims.append(AstChunk(node))


def fstr_candidates(code: str):
    tree = ast.parse(code)

    ch = FstringFinder()
    ch.visit(tree)

    yield from ch.victims


class FstrInliner(ast.NodeTransformer):
    def visit_JoinedStr(self, node):
        new_vals = []
        for v in node.values:
            if (
                isinstance(v, ast.FormattedValue)
                and isinstance(v.value, ast.JoinedStr)
                and v.format_spec is None
            ):
                new_vals += v.value.values
            else:
                new_vals.append(v)

        node.values = new_vals
        return self.generic_visit(node)



import ast
from typing import List

from flynt import state
from flynt.ast_chunk import AstChunk


def is_str_literal(node):
    """ Returns True if a node is a string literal """
    if isinstance(node, ast.Constant):
        return isinstance(node.value, str)
    elif isinstance(node, ast.JoinedStr):
        return True
    else:
        return False


def is_string_concat(node):
    """ Returns True for nodes representing a string concatenation """
    if is_str_literal(node):
        return True
    if isinstance(node, ast.BinOp) and isinstance(node.op, ast.Add):
        return is_string_concat(node.left) or is_string_concat(node.right)
    return False


class ConcatHound(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.victims: List[AstChunk] = []

    def visit_BinOp(self, node: ast.BinOp):
        """
        Finds all nodes that are string concatenations with a literal
        """
        if is_string_concat(node):
            self.victims.append(AstChunk(node))
        else:
            self.generic_visit(node)


def concat_candidates(code: str):
    tree = ast.parse(code)

    ch = ConcatHound()
    ch.visit(tree)

    state.concat_candidates += len(ch.victims)

    yield from ch.victims



import pytest

from flynt.transform.transform import transform_chunk


def test_fmt_spec():
    code = """"my string {:.2f}".format(var)"""
    expected = '''f"""my string {var:.2f}"""'''

    new, changed = transform_chunk(code)

    assert changed
    assert new == expected


def test_expr_no_paren():
    code = """"my string {:.2f}".format(var+1)"""
    expected = '''f"""my string {var + 1:.2f}"""'''

    new, changed = transform_chunk(code)

    assert changed
    assert new == expected


def test_newline():
    code = r""""echo '{}'\n".format(self.FLUSH_CMD)"""
    expected = '''f"""echo '{self.FLUSH_CMD}'\\n"""'''

    new, changed = transform_chunk(code)

    assert changed
    assert new == expected


def test_parenthesis():
    code = """"Flask Documentation ({})".format(version)"""
    expected = '''f"""Flask Documentation ({version})"""'''

    new, changed = transform_chunk(code)

    assert changed
    assert new == expected


def test_implicit_string_concat():
    code = """"Helloo {}" "!!!".format(world)"""
    expected = '''f"""Helloo {world}!!!"""'''

    new, changed = transform_chunk(code)

    assert changed
    assert new == expected


def test_multiline():
    code = """
    "Flask Documentation ({})".format(
    version
)
    """.strip()
    expected = '''f"""Flask Documentation ({version})"""'''

    new, changed = transform_chunk(code)

    assert changed
    assert new == expected


def test_numbered():
    code = '''"""Flask Documentation ({0})""".format(version)'''
    expected = '''f"""Flask Documentation ({version})"""'''

    new, changed = transform_chunk(code)

    assert changed
    assert new == expected


def test_mixed_numbered():
    code = (
        """"Flask Documentation ({1} {0:.2f} {name})".format(version,sprt,name=NAME)"""
    )
    expected = '''f"""Flask Documentation ({sprt} {version:.2f} {NAME})"""'''

    new, changed = transform_chunk(code)

    assert changed
    assert new == expected


def test_unpacking_no_change():
    code = """e.description = "KeyError: '{}'".format(*e.args)"""
    new, changed = transform_chunk(code)
    assert not changed
    assert new == code


def test_kw_unpacking_no_change():
    code = """e.description = "KeyError: '{some_name}'".format(**kwargs)"""
    new, changed = transform_chunk(code)
    assert not changed
    assert new == code


def test_digit_grouping():
    code = """"Failed after {:,}".format(x)"""
    expected = '''f"""Failed after {x:,}"""'''

    new, changed = transform_chunk(code)

    assert changed
    assert new == expected


def test_digit_grouping_2():
    code = """
    "Search: finished in {0:,} ms.".format(vm.search_time_elapsed_ms)
    """.strip()
    expected = '''
    f"""Search: finished in {vm.search_time_elapsed_ms:,} ms."""
    '''.strip()

    new, changed = transform_chunk(code)

    assert changed
    assert new == expected


@pytest.mark.parametrize(
    "s",
    (
        # syntax error
        "(",
        # invalid format strings
        "'{'.format(a)",
        "'}'.format(a)",
        # starargs
        '"{} {}".format(*a)',
        '"{foo} {bar}".format(**b)"',
        # likely makes the format longer
        '"{0} {0}".format(arg)',
        '"{x} {x}".format(arg)',
        '"{x.y} {x.z}".format(arg)',
        # bytestrings don't participate in `.format()` or `f''`
        # but are legal in python 2
        'b"{} {}".format(a, b)',
        # for now, too difficult to rewrite correctly
        '"{a[b]}".format(a=a)',
        '"{a.a[b]}".format(a=a)',
        # not enough placeholders / placeholders missing
        '"{}{}".format(a)',
        '"{a}{b}".format(a=a)',
        # too complex syntax
        '"{:{}}".format(x, y)',
    ),
)
def test_fix_fstrings_noop(s):
    new, changed = transform_chunk(s)
    assert new == s
    assert not changed


@pytest.mark.parametrize(
    ("s", "expected"),
    (
        ('"{} {}".format(a, b)', 'f"""{a} {b}"""'),
        ('"{1} {0}".format(a, b)', 'f"""{b} {a}"""'),
        ('"{x.y}".format(x=z)', 'f"""{z.y}"""'),
        ('"{0.y}".format(z)', 'f"""{z.y}"""'),
        ('"{.y}".format(z)', 'f"""{z.y}"""'),
        ('"{.x} {.y}".format(a, b)', 'f"""{a.x} {b.y}"""'),
        ('"{} {}".format(a.b, c.d)', 'f"""{a.b} {c.d}"""'),
        ('"hello {}!".format(name)', 'f"""hello {name}!"""'),
        ('"{}{{}}{}".format(escaped, y)', 'f"""{escaped}{{}}{y}"""'),
        ('"{}{b}{}".format(a, c, b=b)', 'f"""{a}{b}{c}"""'),
        # TODO: poor man's f-strings?
        # '"{foo}".format(**locals())'
        # TODO: re-evaluate edge cases ported over by #11
        # weird syntax
        ('"{}" . format(x)', 'f"""{x}"""'),
        # spans multiple lines
        ('"{}".format(\n    a,\n)', 'f"""{a}"""'),
    ),
)
def test_fix_fstrings(s, expected):
    new, changed = transform_chunk(s)
    assert changed
    assert new == expected



import re
from pathlib import Path

from setuptools import setup

_DIR = Path(__file__).parent


with (_DIR / "src/flynt/__init__.py").open() as f:
    version = re.search('__version__ = "(.*?)"', f.read()).group(1)


def get_requirements():
    with (_DIR / "requirements.txt").open() as f:
        return f.read()


setup(
    name="flynt",
    packages=["flynt", "flynt.transform", "flynt.lexer", "flynt.string_concat", "flynt.linting"],
    package_dir={"": "src"},
    version=version,
    description="CLI tool to convert a python project's %-formatted strings "
    "to f-strings.",
    author="Ilya Kamenshchikov",
    keywords=["utility", "strings"],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    license="MIT",
    long_description=(_DIR / "README.md").read_text().strip(),
    long_description_content_type="text/markdown",
    install_requires=get_requirements(),
    python_requires=">=3.6",
    entry_points={"console_scripts": ["flynt=flynt:main"]},
    url="https://github.com/ikamensh/flynt",
    include_package_data=True,
)



farewell_message = (
    "Please run your tests before committing. Did flynt get a perfect conversion? give it a star at: "
    "\n~ https://github.com/ikamensh/flynt ~"
    "\nThank you for using flynt. Upgrade more projects and recommend it to your colleagues!\n"
)



import random

import pytest

from flynt.format import QuoteTypes, get_quote_type, set_quote_type
from flynt.lexer.split import get_chunks


@pytest.mark.parametrize(
    argnames=["code", "quote_type"],
    argvalues=[
        ("'abra'", QuoteTypes.single),
        ('"bobro"', QuoteTypes.double),
        ("'''abra'''", QuoteTypes.triple_single),
        ('"""bobro"""', QuoteTypes.triple_double),
    ],
)
def test_get_quote_type_token(code, quote_type):

    g = get_chunks(code)
    next(g)
    chunk = next(g)
    token = chunk.tokens[0]

    assert token.get_quote_type() == quote_type


@pytest.mark.parametrize(
    argnames=["code", "quote_type"],
    argvalues=[
        ("'abra'", QuoteTypes.single),
        ('"bobro"', QuoteTypes.double),
        ("'''abra'''", QuoteTypes.triple_single),
        ('"""bobro"""', QuoteTypes.triple_double),
    ],
)
def test_get_quote_type(code, quote_type):
    assert get_quote_type(code) == quote_type


@pytest.mark.parametrize(
    argnames="code", argvalues=["'abra'", '"bobro"', "'''abra'''", '"""bobro"""']
)
def test_cycle(code):
    assert set_quote_type(code, get_quote_type(code)) == code


@pytest.mark.parametrize(argnames="quote_type", argvalues=QuoteTypes.all)
def test_initial_doesnt_matter(quote_type):
    code = random.choice(["'abra'", '"bobro"', "'''abra'''", '"""bobro"""'])
    assert get_quote_type(set_quote_type(code, quote_type)) == quote_type


def test_single():
    code = '"alpha123"'
    expected = "'alpha123'"

    assert set_quote_type(code, QuoteTypes.single) == expected


def test_single_from_triple():
    code = '"""alpha123"""'
    expected = "'alpha123'"

    assert set_quote_type(code, QuoteTypes.single) == expected



from flynt.lexer import split


def test_str_newline():
    s_in = """a = '%s\\n' % var"""
    generator = split.get_fstringify_chunks(s_in)
    assert len(list(generator)) == 1


def test_triple():
    s_in = """print("{}".format(Bar + 1), '%d' % var, "{s}".format(s=foo))"""
    generator = split.get_fstringify_chunks(s_in)
    assert len(list(generator)) == 3


def test_one_string():
    s = """"my string {}, but also {} and {}".format(var, f, cada_bra)"""
    chunks_gen = split.get_fstringify_chunks(s)
    assert len(list(chunks_gen)) == 1

    generator = split.get_fstringify_chunks(s)
    chunk = next(generator)

    assert chunk.start_line == 0
    assert s[: chunk.end_idx] == s


def test_yields_parsable():
    code_in = """attrs = {'r': '{}'.format(row_idx)}"""
    generator = split.get_fstringify_chunks(code_in)
    chunk = next(generator)

    assert chunk.is_parseable
    assert code_in[chunk.start_idx : chunk.end_idx] == "'{}'.format(row_idx)"


def test_percent_attribute():
    code_in = """src_info = 'application "%s"' % srcobj.import_name"""

    generator = split.get_fstringify_chunks(code_in)
    chunk = next(generator)

    expected = """'application "%s"' % srcobj.import_name"""
    assert code_in[chunk.start_idx : chunk.end_idx] == expected


def test_percent_call():
    code_in = """"filename*": "UTF-8''%s" % url_quote(attachment_filename)"""

    generator = split.get_fstringify_chunks(code_in)
    chunk = next(generator)

    expected = """"UTF-8''%s" % url_quote(attachment_filename)"""
    assert code_in[chunk.start_idx : chunk.end_idx] == expected


def test_two_strings():
    s = (
        'a = "my string {}, but also {} and {}".format(var, f, cada_bra)\n'
        + 'b = "my string {}, but also {} and {}".format(var, what, cada_bra)'
    )

    chunks_gen = split.get_fstringify_chunks(s)
    assert len(list(chunks_gen)) == 2

    generator = split.get_fstringify_chunks(s)
    lines = s.split("\n")

    chunk = next(generator)

    assert chunk.start_line == 0
    assert lines[0][: chunk.end_idx] == lines[0]

    chunk = next(generator)

    assert chunk.start_line == 1
    assert lines[1][: chunk.end_idx] == lines[1]


def test_indented():
    indented = """
    var = 5
    if var % 3 == 0:
        a = "my string {}".format(var)""".strip()

    generator = split.get_fstringify_chunks(indented)
    assert len(list(generator)) == 1
    lines = indented.split("\n")

    generator = split.get_fstringify_chunks(indented)
    chunk = next(generator)

    assert chunk.start_line == 2
    assert lines[2][: chunk.end_idx] == lines[2]


def test_empty_line():
    code_empty_line = """
    def write_row(self, xf, row, row_idx):
    
        attrs = {'r': '{}'.format(row_idx)}""".strip()

    generator = split.get_fstringify_chunks(code_empty_line)
    lines = code_empty_line.split("\n")

    chunk = next(generator)

    assert chunk.start_line == 2
    assert lines[2][chunk.start_idx : chunk.end_idx] == "'{}'.format(row_idx)"


multiline_code = """
raise NoAppException(
            'Detected multiple Flask applications in module "{module}". Use '
            '"FLASK_APP={module}:name" to specify the correct '
            "one.".format(module=module.__name__)
        )
""".strip()


def test_multiline():
    generator = split.get_fstringify_chunks(multiline_code)
    assert len(list(generator)) == 1


not_implicit_concat = """
html_logo = "_static/flask-logo-sidebar.png"
html_title = "Flask Documentation ({})".format(version)""".strip()


def test_not_implicit_concat():
    generator = split.get_fstringify_chunks(not_implicit_concat)
    assert len(list(generator)) == 1


line_continuation = """
a = "Hello {}" \\
"world".format(',')""".strip()


def test_line_continuation():
    generator = split.get_fstringify_chunks(line_continuation)
    assert len(list(generator)) == 1


tuple_in_list = """
latex_documents = [
    (master_doc, "Flask-{}.tex".format(version), html_title, author, "manual")
]""".strip()


def test_tuple_list():
    generator = split.get_fstringify_chunks(tuple_in_list)
    assert len(list(generator)) == 1


def test_indexed_percent():
    code = 'return "Hello %s!" % flask.request.args[name]'
    generator = split.get_fstringify_chunks(code)
    chunk = next(generator)

    assert (
        code[chunk.start_idx : chunk.end_idx]
        == '"Hello %s!" % flask.request.args[name]'
    )


def test_tuple_percent():
    code = """print("%s %s " % (var+var, abc))"""
    generator = split.get_fstringify_chunks(tuple_in_list)
    assert len(list(generator)) == 1
